
/**
 *@file mod_countm.c
 *@brief countm module routines
 *@author Byron Young
 *@remark A simple apache2 DOS module page counter returning
 *@remark jpeg or png images.
 *@version 3.0
 *@bug report to spamiccling@yahoo.com
 *@bug report to bkyoung@users.sourceforge.net
 */

/**
 *@package apache2-mod_countm
 */
#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <apr.h> /* apr_uint32_t */
#include <apr_strings.h>
#include <apr_lib.h>
#include <apr_file_io.h>
#include <apr_file_info.h>
#include <apr_pools.h>
#include <http_protocol.h>

#include <gd.h>
#include <db.h>

#include <string.h>
#include <ctype.h>
#include <unistd.h>

/**
 *@def DEFAULT_LOCKFILE_STRING
 *@brief Default countm lockfile.
 *When file exists, indicates database is locked, and in use.
 *Define as an absolute path/filename.
 *May not be the empty string.
 *@warning if file exists when module loaded an error is returned.
 */
#define DEFAULT_LOCKFILE_STRING "/var/countm/countm.lock"

/**
 *@def DEFAULT_FONTSDIR_STRING
 *@brief Default countm fonts directory.
 *Specifies the directory containing the true type fonts countm may use.
 *Must be and absolute path.
 *May not be the empty string (which is an error).
 *Finally, a '/' must not end the string. 
*/
#define DEFAULT_FONTSDIR_STRING "/var/countm/fonts"

/**
 *@def DEFAULT_FONT_STRING
 *@brief Defalut countm font.
 *Must be defined as a valid filename.
 *Do not include the .ttf extension suffix.
 *Is considered relative to DEFAULT_FONTSDIR_STRING.
 *A '/' should not begin the name.
 *@warning The font file must be in fontsdir, or, when module is loaded,
 *an error is returned.
*/
#define DEFAULT_FONT_STRING "FreeMono"

/**
 *@def DEFAULT_DB_ENVIRONMENT_HOME_STRING
 *@brief The directory containing database infomation.
 *Must be absolute.
 *A '/' should not end the path.
 */
#define DEFAULT_DB_ENVIRONMENT_HOME_STRING "/var/countm/dbase"

/**
 *@def DEFAULT_DB_FILE_STRING
 *@brief The database file of link access hostnames.
 *Relative to the DEFAULT_DB_ENVIRONMENT_HOME_STRING directory.
 *May not be empty string.
 */
#define DEFAULT_DB_FILE_STRING "countm.db"

/**
 *@def DEFAULT_DB_LOG_FILE_STRING
 *@brief The database log file.
 *Must be absolute.
 *May not be empty string.
 */
#define DEFAULT_DB_LOG_FILE_STRING "/var/countm/dbase/db.log"

/**
 *@def DB_ERROR_LOG_PREFIX_STRING
 *@brief The prefix used when writing database errors.
 */
#define  DB_ERROR_LOG_PREFIX_STRING "dba_err"

/**
 *@def DB_COUNTER
 *@brief The internal name of the counter database.
 */
#define DB_COUNTER "DB_COUNTER"

/**
 *@def DB_ACCESS
 *@brief The internal name of the access database.
 */
#define DB_ACCESS "DB_ACCESS"

/**
 *@def MODULE_URI_PATH_STRING
 *@brief The match string for the URI path.
 *Must begin with a /, and be a valid path name.
 *Only alter if absolutely necessary.
 *Matches the directory specified in the <Location> directive.
 */
#define MODULE_URI_PATH_STRING "/countm"

/**
 *@def MODULE_HANDLER_ID_STRING
 *@brief The AddHandler string in the config file
 *Matches the handler string to the module.
 *Only alter if absolutely necessary.
 */
#define MODULE_HANDLER_ID_STRING "mod-countm"

/**
 *@def USE_REMOTE_DOUBLE_REV
 *@brief If hostname lookup fails, attempts to use double reverse
 *If get_remote_host fails to return a valid hostname,
 *an additional call using REMOTE_DOUBLE_REV will be attempted.
 *To deactivate, comment out the define.
 */
#define USE_REMOTE_DOUBLE_REV

/**
 *@def USE_PNG_IF_ACCEPTABLE
 *@brief Use png if remote client accepts it.
 *If the remote client request a jpeg image, but can accept a png image,
 *return a png image instead of jpeg image.
 */
/*#define USE_PNG_IF_ACCEPTABLE*/

/**
 *@def UNLOCK_ATTEMPTS
 *@brief The number of database unlock attempts each request tries.
 *If the database may not be unlocked,
 *the request sleeps for a second and tries again.
 *After each attempt, it waits for 1 second.
 */
#define UNLOCK_ATTEMPTS 3

/**
 *@def INSERT_BUFFER_SIZE
 *@brief The size, in bytes, of the insert buffer.
 *As the commands are parsed off the uri, all insert=
 *values are added into the buffer.
 *The buffer is terminated with a double null.
 *If adding the insert=value would overflow the buffer,
 *the insert=value is not added, and is discarded.
 */
#define INSERT_BUFFER_SIZE 1024

/**
 *@def MAX_WIDTH_CHAR_DISPLAY
 *@brief The maximum number of characters/digits to return in image.
 */
#define MAX_WIDTH_CHAR_DISPLAY 10

/**
 *@def DEFAULT_FONT_POINT_SIZE
 *@brief The default font point size.
 */
#define DEFAULT_FONT_POINT_SIZE 20

/**
 *@def DEFAULT_IMAGE_TYPE
 *@brief The default image type.
 *one of img_jpeg, img_png
 */
#define DEFAULT_IMAGE_TYPE img_jpeg

/**
 *@def TEMPORARY_WRITEABLE_DIRECTORY_PATH
 *@brief Path to a writeable directory for temporary files.
 *Do not include a trailing slash.
 *Include quotes.
 *Only required for jpeg images.
 */
#define TEMPORARY_WRITEABLE_DIRECTORY_PATH "/var/tmp"

#define COUNTM_INFORMATION_PRIORITY_DATABASE     1
#define COUNTM_INFORMATION_PRIORITY_COMMANDLINE  2

/**
 *@def COUNTM_INFORMATION_PRIORITY
 *@brief Determines if database or command line information has priority
 *When info from the database conflicts with info from command line,
 *use this to determine priority.
 *COUNTM_INFORMATION_PRIORITY_DATABASE = database information has priority
 *COUNTM_INFORMATION_PRIORITY_COMMANDLINE = commandline information has priority
 */
#define COUNTM_INFORMATION_PRIORITY COUNTM_INFORMATION_PRIORITY_DATABASE

/**
 *@def MAX_FONT_NAME_SIZE
 *@brief The maximum chars in a font name.
 *The maximum chars in a font name. Does not include the suffix .ttf,
 *nor any leading path characters.
 *Used for storage in DB_COUNT.
 */
#define MAX_FONT_NAME_SIZE 128

/* misc constants */
#define MAX_CSTRING 256
#define MAX_QSTRING 1024

#ifdef DEBUG
#define debug_out(fp, fmt, ...) fprintf(fp, fmt, __VA_ARGS__)
#else
#define debug_out(fp, fmt, ...)
#endif

/**
 *@enum imagetype
 *@brief The image type enumeration
 */
enum imagetype
{
  img_notdef=-1, /**< image type not defined */
  img_jpeg=0, /**< image type specified as jpeg */
  img_png=1 /**< image type specified as png */
};

/**
 *@enum randomtype
 *@brief The random type enumeration
 */
enum randomtype
{
  random_notdef=-1, /**< random type not defined */
  random_true=0, /**< return a random value, remote host is not inserted into DB_ACCESS */
  random_false=1, /**< not a random value specifically indicated */
  random_true_access=2 /**< return a random value, but insert remote host into DB_ACCESS */
};

/**
 *@enum resettype
 *@brief The reset type enumeration
 */
enum resettype
{
  reset_notdef, /**< reset type not defined */
  reset_value, /**< reset type defined as a value */
  reset_delete, /**< reset type specified as delete */
  reset_init, /**< reset type specified as init */
  reset_access /**< reset type specified as access */
};

/**
 *@enum
 *@brief The mode type enumeration
 */
enum counttype
{
  count_notdef, /**< mode type not defined */
  count_inc, /**< count type specified as increment */
  count_dec, /**< count type specified as decrement */
  count_norm_inc, /**< count type sepecifed as normal increment */
  count_norm_dec /**< count type specified as normal decrement */
};

/**
 *@brief The type for the counter value
 */
typedef unsigned long cntval_t;

/**
 *@brief Web color identifier
 */
struct rgb
{
  int rgb_r;
  int rgb_g;
  int rgb_b;
};

/**
 *@struct db_count_record
 *@brief the structure stored in DB_COUNT
 *The structure that is stored in DB_COUNT.
 *The values that may be set as default during
 *the initialization of counter.
 */
struct db_count_record
{
  cntval_t cnt;
  char font[MAX_FONT_NAME_SIZE]; /**< name of font to use. '\0'=>not specified */
  enum imagetype imgtype; /**< the stored default image type */
  enum counttype count; /**< the stored default count type */
  enum randomtype random; /**< the stored default random type */
  unsigned int status; /**< bit field for various informations
  * bit 0 = width_val defined, bit 1=point_val defined, bit 2=bgcolor_val defined, bit 3=text_val defined
  */
  int width_val; /**< the stored default chars/digits to return */
  int point_val; /**< the stored default point size of font */
  struct rgb bgcolor_val; /**< the stored default background color */
  struct rgb text_val; /**< the stored default text color */
};

/**
 *@struct countent
 *@brief the structure that stores all the info
 *The structure that stores all the info parsed from the command
 *line, and used to display the image.
 */
struct countent
{
  const char * ce_link; /**< the link name, only specified once, NULL=not specified */
  enum resettype ce_reset; /**< see enum resettype */
  cntval_t ce_reset_val; /**< if given, the value to reset counter to */
  int ce_width; /**< 0=not specified, 1= specified */
  unsigned long ce_width_val; /**< if specified, this is the width value */
  char * ce_font; /**< if specified, the font name. NULL=not specified */
  int ce_ignore; /**< 0=ignore not given, or not matched. 1=ignore matched */
  int ce_bgcolor; /**< background color 0=not specified, 1=specified */
  struct rgb ce_bgcolor_val; /**< if specified, value */
  int ce_text; /**< 0 = not specified, 1=specified */
  struct rgb ce_text_val; /**< if specified, value */
  int ce_point; /**< 0=not specified, 1=specified */
  int ce_point_val; /**< if specified, value */
  enum imagetype ce_imgtype; /**< -1=not given, 0=jpeg (default), 1=png */
  enum counttype ce_count; /**< the counter count type */
  enum randomtype ce_random; /**< the random information */
  int ce_insert; /**< 1= was given, and insert_buffer not empty, 0 = not given */
  char ce_insert_buffer[INSERT_BUFFER_SIZE]; /**< the insert= command buffer */

  unsigned long ce_error; /**< 0=no error, >1 = error happened, and the error is in ce_string */
  const char * ce_hostname; /**< remote hostname  */
  const char * ce_ipaddr; /**< remote  ipaddr */
  int ce_hostname_is_ipaddr; /**< non zero if ce_ipaddr is and address */

  cntval_t ce_db_val; /**< the value of count from db, after merging , or after reading */
  int ce_value; /**< 0=no value, 1=value to convert to string in ce_value_val */
  unsigned long ce_value_val; /**< value to convert to string if ce_value==1 */
  char * ce_string; /**< the string to display */

  /* dbase stuff */
  u_int32_t ce_cachesize; /**< size of db4 cachesize */
  FILE * ce_fp_log; /**< db4 log file */
  DB_ENV * ce_dbenv; /**< db4 environment */
  DB * ce_dbp_a; /**< DB_ACCESS database pointer */
  DB * ce_dbp_c; /**< DB_COUNT database pointer */
  int ce_db_log_file; /**< db4 logfile state 0=closed/not open, 1=opened */
  int ce_link_in_db_count; /**< Is link in DB_COUNT? 0=no, 1=yes */
  int ce_link_in_db_access; /**< Is link in DB_ACCESS? 0=no, 1=yes */
  struct db_count_record ce_dbc_rec; /**< link info from DB_COUNT */
};

/* The initial values prior to parsing command line */
#define INITIALIZE_COUNTENT {\
    .ce_link = NULL,\
    .ce_reset = reset_notdef,\
    .ce_reset_val=0,\
    .ce_width = 0,\
    .ce_width_val=0,\
    .ce_font=NULL,\
    .ce_ignore=0,\
    .ce_bgcolor=0,\
    .ce_bgcolor_val={0,0,0},\
    .ce_text=0,\
    .ce_text_val={255,255,255},\
    .ce_point=0,\
    .ce_point_val=DEFAULT_FONT_POINT_SIZE,\
    .ce_imgtype=img_notdef,\
    .ce_count=count_norm_inc,\
    .ce_random=0,\
    .ce_insert=0,\
    .ce_insert_buffer="\0\0",\
    .ce_error=0,\
    .ce_hostname=NULL,\
    .ce_ipaddr=NULL,\
    .ce_hostname_is_ipaddr=0,\
    .ce_string=NULL\
}

/**
 *@struct countm_config
 *@brief per-server-configuration information
 *Presently, not used.
 */
struct countm_config
{
  int bConfigError;   /**<  0=Module configuration correct.
			* !0=Module configuration error.
			* @brief per-server-configuration status code
			*/

};

extern const char * const version;

static const char sza_def_lockfile[]=DEFAULT_LOCKFILE_STRING;
static const char sza_def_fontsdir[]=DEFAULT_FONTSDIR_STRING;
static const char sza_def_font[]=DEFAULT_FONT_STRING;
static const char sza_def_db_environment_home[]=
                                DEFAULT_DB_ENVIRONMENT_HOME_STRING;
static const char sza_def_db_file[]=DEFAULT_DB_FILE_STRING;
static const char sza_def_db_log_file[]=DEFAULT_DB_LOG_FILE_STRING;
static const char sza_def_db_error_log_prefix[]=DB_ERROR_LOG_PREFIX_STRING;

static const char * sza_lockfile=sza_def_lockfile;
static const char * sza_fontsdir=sza_def_fontsdir;
static const char * sza_font=sza_def_font;
static const char * sza_db_environment_home = sza_def_db_environment_home;
static const char * sza_db_file=sza_def_db_file; /* counter_db */
static const char * sza_db_log_file = sza_def_db_log_file;
static const char * sza_db_error_log_prefix=sza_def_db_error_log_prefix;

static void * countm_per_server_config(apr_pool_t *p, server_rec *s);
static int countm_setdefaults(request_rec *r, struct countent * p_ce);
static int countm_parse_cmdline(request_rec *r, struct countent * p_ce);
static int countm_method_handler(request_rec *r);
static void countm_register_hooks(apr_pool_t *p);
static int countm_lock(const char * const sza_fn, apr_pool_t * p);
static int countm_unlock(const char * const sza_fn, apr_pool_t * p);
static const char * countm_set_lockfile(cmd_parms *parms, void *dummy, const char *arg);
static const char * countm_set_fontsdir(cmd_parms *parms, void *dummy,
				 const char *arg);
static const char * countm_set_font(cmd_parms *parms, void *dummy, const char *arg);
static const char * countm_set_db_environment_home(cmd_parms *parms, void *dummy,
					    const char *arg);
static const char * countm_set_db_file(cmd_parms *parms, void *dummy,
				const char *arg);
static const char * countm_set_db_log_file(cmd_parms *parms, void *dummy,
				    const char *arg);
static void countm_rgb_color(const char *hexnum, struct rgb *rgb);
static int countm_apmod_sink(void *r, const char *buf, int len);
static int countm_display_string(struct countent * ce, request_rec *r);
#ifdef DEBUG
static void dbg_fout(apr_file_t * fp, char * buf);
#endif
static DB_ENV * countm_db_env_init(const char * home, const char * prefix,
			    int cachesize, FILE * fp_log);
static DB * countm_db_open(DB_ENV * dbenv, const char * fname, const char * dname,
		    DBTYPE dtype, u_int32_t dflags_open, u_int32_t dflags_db,
		    int dmode, u_int32_t dpgsize);
static int countm_misc_dza_strcat( char * const buf, char const * const szstr, const int bsize);
#ifdef DEBUG
void countm_debug_print_countent(FILE * fp, struct countent * p_ce);
void countm_debug_print_db_count_record(FILE * fp, struct db_count_record * p);
void countm_debug_print_request_rec(FILE * fp, request_rec * r);
#endif
static int countm_db_open_databases(struct countent * p_ce, request_rec * r);
static void countm_db_close_databases(struct countent * p_ce);
static int countm_db_count_get_link(struct countent * p_ce);
static int countm_merge_info(struct countent * p_ce, request_rec * r);
static int countm_handler_reset(struct countent * p_ce, request_rec *r);
static int countm_fill_db_count_rec(struct countent * p_ce);
static int countm_handler_random(struct countent * p_ce, request_rec *r);
static int countm_handler_updatecounter(struct countent * p_ce, request_rec *r);


#ifdef DEBUG
static FILE * dbg_log=NULL;
#define DEBUG_LOG_FILE_STRING "/var/countm/dbg.dlog"
#endif

/**
 *@struct countm_cmds
 *@brief Defines the directives used in the configuration file.
 *@param CountmDB The name of the countm database file.
 *@param CountmFontDir The name of the countm font directory.
 *@param CountmPrefix The name of the countm config files directory.
 *@param CountmLockfile The name of the countm database lockfile.
 *@param CountmDefFont The name of the default countm font.
 */
command_rec countm_cmds[]=
{
  AP_INIT_TAKE1("CountmLockfile", countm_set_lockfile, NULL, RSRC_CONF,
		"the name of the countm database lockfile"),

  AP_INIT_TAKE1("CountmFontDir", countm_set_fontsdir, NULL, RSRC_CONF,
		"the name of the countm font directory"),

  AP_INIT_TAKE1("CountmDefFont", countm_set_font, NULL, RSRC_CONF,
		"the name of the default countm font"),

  AP_INIT_TAKE1("CountmEnvironmentHome", countm_set_db_environment_home, NULL,
		RSRC_CONF, "the name of the countm database directory"),

  AP_INIT_TAKE1("CountmDB", countm_set_db_file, NULL, RSRC_CONF,
		"the name of the countm database file"),
  
  AP_INIT_TAKE1("CountmDBLog", countm_set_db_log_file, NULL, RSRC_CONF,
		"the name of the countm database log file"),
  
  {NULL}
};


module  AP_MODULE_DECLARE_DATA countm_module =
  {
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    countm_per_server_config,
    NULL,
    countm_cmds,
    countm_register_hooks
  };

/**
 *@fn countm_per_server_config(apr_pool_t *p, server_rec *s)
 *@brief Server startup configuration routine.
 *@param p The configuration pool. Exists till server shutdown.
 *@param s The server record.
 *@return Pointer to the module_config record. What happens if return NULL?
 */

/*
 *For now, does nothing.
 */
static void *
countm_per_server_config(apr_pool_t *p, server_rec *s)
{
  struct countm_config * p_oc;
  struct countm_config def_oc=
    {
      1    /* assume fail */
    };

  p_oc = (struct countm_config *)
         apr_pcalloc(p, sizeof(struct countm_config));
  if (p_oc ==  NULL)
    {
      return NULL;
    }
  memcpy(p_oc, &def_oc, sizeof(struct countm_config));
  return (void *) p_oc;
}

/**
 *@fn countm_hook_post_config(apr_pool_t *pconf, apr_pool_t * plog,
 *@fn                         apr_pool_t *ptemp, server_rec *s)
 *@brief The post configuation hook.
 *Assigns default values to parameters not given in the httpd.conf file,
 *and verifies that required files exist, and are accessable.
 *How return failure, and abort module loading?
 *@param pconf The configuration pool.
 *@param plog The log pool (whatever that is)??.
 *@param ptemp The temporary pool. Created at the beginning of configuration
 *cycle, and exists until the end of config parsing (same as cmd->temp_pool).
 *@param s Server record.
 *@return Message code. What are the possible values?
 */
static int
countm_hook_post_config(apr_pool_t *pconf, apr_pool_t * plog,
			  apr_pool_t *ptemp, server_rec *s)
{
  apr_file_t * fp;
  apr_status_t fstatus=0;
  char * db_environment_home;
  char * db_file;
  char * fontsdir;
  char * pc;
  struct countm_config * p_oc;

  /* verify lockfile can be created */
  fstatus=apr_file_open(&fp, sza_lockfile, APR_READ | APR_WRITE | APR_CREATE |
			APR_EXCL | APR_DELONCLOSE, APR_OS_DEFAULT, pconf);
  if (fstatus!=APR_SUCCESS)
    {
      /* file already exists, or some other error */
      /* logit and fail*/
      ap_log_perror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, pconf,
		    "Open Failed: %s", sza_lockfile);
    }
  else
    {
      apr_file_close(fp);
    }

  /* verify default font may be opened */
  fontsdir = apr_pstrdup(pconf, sza_fontsdir);
  fontsdir = apr_pstrcat(pconf, fontsdir, "/", NULL);
  pc = apr_pstrcat(pconf, fontsdir, sza_font, ".ttf", NULL);
  fstatus = apr_file_open(&fp, pc, APR_READ, APR_OS_DEFAULT, pconf);
  if (fstatus!=APR_SUCCESS)
    {
      /* cannot open default font file for reading */
      /* logit and fail */
      ap_log_perror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, pconf,
		    "Open Failed: %s", pc);
    }
  else
    {
      apr_file_close(fp);
    }
  sza_fontsdir = fontsdir;

  /* verify sza_db_environment_home/sza_db_file */
  db_environment_home = apr_pstrdup(pconf, sza_db_environment_home);
  db_environment_home = apr_pstrcat(pconf, db_environment_home, "/", NULL);

  db_file = apr_pstrdup(pconf, db_environment_home);
  db_file = apr_pstrcat(pconf, db_file, sza_db_file, NULL);
  fstatus=apr_file_open(&fp, db_file, APR_READ | APR_WRITE ,
			APR_OS_DEFAULT, pconf);
  if (fstatus==APR_SUCCESS)
    {
      apr_file_close(fp);
    }
  else
    {
      fstatus = apr_file_open(&fp, db_file, APR_READ | APR_WRITE | APR_CREATE,
			      APR_OS_DEFAULT, pconf);
      if (fstatus == APR_SUCCESS)
	{
	  apr_file_close(fp);
	  fstatus = apr_file_remove(db_file, pconf);
	  if (fstatus != APR_SUCCESS)
	    {
	      ap_log_perror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, pconf,
			    "Could not remove %s", db_file);
	    }
	}
      else
	{
	  /* cannot access database file */
	  /* logit and fail*/
	  ap_log_perror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, pconf,
			"Cannot access db file: %s", db_file);
	}
    }
  sza_db_file = db_file;
  sza_db_environment_home = db_environment_home;
  /* now, sza_db_file is the actual database path/filename string */
  /* sza_db_environment_home is the actual directory path */

  /* verify and start a new sza_db_log_file */
  pc = (char *) sza_db_log_file;
  fstatus = apr_file_open(&fp, pc,
			  APR_READ | APR_WRITE | APR_CREATE | APR_DELONCLOSE,
			  APR_OS_DEFAULT, pconf);
  if (fstatus == APR_SUCCESS)
    {
      apr_file_close(fp);
    }
  else
    {
      /* cannot access database log file */
      /* logit and fail*/
      ap_log_perror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, pconf,
		    "Cannot access db log file: %s", pc);
    }

#ifdef DEBUG
  fstatus=apr_file_open(&fp, "/var/countm/countm.dlog",
			APR_READ | APR_WRITE | APR_CREATE,
			APR_OS_DEFAULT, pconf);
  if (fstatus!=APR_SUCCESS)
    {
      /* file already exists, or some other error */
      /* logit and fail*/
      ap_log_perror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, pconf,
		    "Open debug file failed");
    }
  else
    {
      dbg_fout(fp, "sza_lockfile:");
      dbg_fout(fp, (char *) sza_lockfile);
      dbg_fout(fp, "sza_fontsdir:");
      dbg_fout(fp, (char *) sza_fontsdir);
      dbg_fout(fp, "sza_font:");
      dbg_fout(fp, (char *) sza_font);
      dbg_fout(fp, "sza_db_environment_home:");
      dbg_fout(fp, (char *) sza_db_environment_home);
      dbg_fout(fp, "sza_db_file:");
      dbg_fout(fp, (char *) sza_db_file);
      dbg_fout(fp, "sza_db_log_file:");
      dbg_fout(fp, (char *) sza_db_log_file);

      apr_file_close(fp);
    }
#endif
  p_oc = (struct countm_config *) ap_get_module_config(s->module_config,
						       &countm_module);
  p_oc->bConfigError=0;
  return OK;
}

#ifdef DEBUG
static void
dbg_fout(apr_file_t * fp, char * buf)
{
  apr_size_t nbytes;

  nbytes=strlen(buf);
  apr_file_write(fp, buf, &nbytes);
  nbytes=1;
  apr_file_write(fp, "\n", &nbytes);
  return;
}
#endif

/**
 *@fn countm_register_hooks(apr_pool_t *p)
 *@brief Registers required hooks.
 *Called by API to register hooks necessary for various stages.
 *@param p The memory pool. The pconf, ptemp, or other pool?
 *@return none
 */
static void
countm_register_hooks(apr_pool_t *p)
{
  ap_hook_post_config(countm_hook_post_config, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_handler(countm_method_handler, NULL, NULL, APR_HOOK_LAST);
  return;
}

/**
 *@fn countm_setdefaults(request_rec *r, struct countent * p_ce)
 *@brief Set any undefined value to a default value.
 *Process the information, and verify all values not defined
 *are set to default values.
 *@return Always returns 0
 */
static int
countm_setdefaults(request_rec *r, struct countent * p_ce)
{
  const struct rgb def_bg={0,0,0};
  const struct rgb def_text={255,255,255};  

  if (p_ce->ce_width == 0)
    {
      p_ce->ce_width_val=0;
      p_ce->ce_width=1;
    }
  /* default font */
  if (p_ce->ce_font == NULL)
    {
      p_ce->ce_font = apr_pstrdup(r->pool, sza_font);
    }
  if (p_ce->ce_bgcolor == 0)
    {
      memcpy(&p_ce->ce_bgcolor_val,&def_bg,sizeof(def_bg));
      p_ce->ce_bgcolor=1;
    }
  if (p_ce->ce_text == 0)
    {
      memcpy(&p_ce->ce_text_val,&def_text,sizeof(def_text));
      p_ce->ce_text=1;
    }
  if (p_ce->ce_point == 0)
    {
      p_ce->ce_point_val=DEFAULT_FONT_POINT_SIZE;
      p_ce->ce_point=1;
    }
  /* default image type */
  if (p_ce->ce_imgtype==img_notdef)
    {
      p_ce->ce_imgtype=DEFAULT_IMAGE_TYPE;
    }
  /* default string to display */
  if (p_ce->ce_string==NULL)
    {
      p_ce->ce_string="Estr";
    }
  if(p_ce->ce_count == count_notdef)
    {
      p_ce->ce_count=count_norm_inc;
    }
  return 0;
} /* countm_setdefaults */

/**
 *@fn countm_parse_cmdline(request_rec *r, struct countent * p_ce)
 *@brief Parses the URI.
 *Processes the request given in r->args into parameters,
 *@param r The request record of the method.
 *@param p_ce The countent information.
 * All errors returned are query ending errors. The database will not be accessed,
 * and default values will be used to return the image (if possible).
 *@return Status value.
 *  0 = SUCCESS. Command line parsed.
 *  1 = ERROR. Hostname or ipaddr is NULL.
 *  2 = ERROR. Post method error.
 *  3 = ERROR. Some type of parameter format error.
 *  4 = ERROR. An invalid font was given.
 *  5 = ERROR. Could not determine a valid link name.
 *
 */
#define ERROR_PARSE_CONNECTION 1
#define ERROR_PARSE_METHOD     2
#define ERROR_PARSE_PARAMETER  3
#define ERROR_PARSE_FONT       4
#define ERROR_PARSE_LINK       5
static int
countm_parse_cmdline(request_rec *r, struct countent * p_ce)
{
  const char *data, *param;
  const char *re=NULL, *p=NULL, *ig=NULL, *ra=NULL;
  const char *wi=NULL, *fo=NULL, *bg=NULL, *tx=NULL;
  const char *pt=NULL, *it=NULL, *ct=NULL;
  char *re_init=NULL;
  const struct countent ce_init=INITIALIZE_COUNTENT;
  apr_status_t fstatus;
  apr_file_t * fp;
  int ret;
  /* clear out insert buffer for insert= list */
  memcpy(p_ce, &ce_init, sizeof(struct countent));
  p_ce->ce_insert_buffer[0]='\0';
  p_ce->ce_insert_buffer[1]='\0';
  p_ce->ce_link=NULL;
  p_ce->ce_reset=reset_notdef;
  p_ce->ce_reset=0;
  p_ce->ce_width=0;
  p_ce->ce_width_val=0;
  p_ce->ce_font=NULL;
  p_ce->ce_ignore=0;
  p_ce->ce_bgcolor=0;
  p_ce->ce_bgcolor_val.rgb_r=0;
  p_ce->ce_bgcolor_val.rgb_g=0;
  p_ce->ce_bgcolor_val.rgb_b=0;
  p_ce->ce_text=0;
  p_ce->ce_text_val.rgb_r=0;
  p_ce->ce_text_val.rgb_g=0;
  p_ce->ce_text_val.rgb_b=0;
  p_ce->ce_point=0;
  p_ce->ce_imgtype=img_notdef;
  p_ce->ce_count=count_notdef;
  p_ce->ce_random=random_notdef;
  p_ce->ce_insert=0;
  p_ce->ce_error=0;

  /* be certain HostNameLookups in httpd.conf on!! */
  p_ce->ce_hostname = ap_get_remote_host(r->connection, r->per_dir_config,
				REMOTE_NAME, &p_ce->ce_hostname_is_ipaddr);

#ifdef USE_REMOTE_DOUBLE_REV
  if (p_ce->ce_hostname_is_ipaddr || (p_ce->ce_hostname==NULL))
    {
      if (!r->connection->double_reverse)
	p_ce->ce_hostname = ap_get_remote_host(r->connection, r->per_dir_config,
					       REMOTE_DOUBLE_REV, NULL);
    }
#endif

  p_ce->ce_ipaddr=r->connection->remote_ip;
  if (p_ce->ce_hostname==NULL || p_ce->ce_ipaddr==NULL)
    {
      p_ce->ce_error=ERROR_PARSE_CONNECTION;
      p_ce->ce_string = "EConn";
      return ERROR_PARSE_CONNECTION;
    }
  if (r->method_number==M_POST)
    {
      p_ce->ce_error=ERROR_PARSE_METHOD;
      p_ce->ce_string = "EPost";
      return ERROR_PARSE_METHOD;
    }

  data = r->parsed_uri.query;
  while ( data && *data && ( param = ap_getword( r->pool, &data, '&' )))
    {
    if (( strlen( data ) > MAX_QSTRING ) || ( param == NULL ))
      {
	p_ce->ce_error=ERROR_PARSE_PARAMETER;
	p_ce->ce_string = "EParm";
	return ERROR_PARSE_PARAMETER;
      }
    p = ap_getword( r->pool, &param, '=' );
    if ( p == NULL )
      {
	p_ce->ce_error=ERROR_PARSE_PARAMETER;
	p_ce->ce_string = "EParm";
	return ERROR_PARSE_PARAMETER;
      }
    if ( strcasecmp( p, "link") == 0 )
      {
	/* check to see if other link= specifications */
	if (p_ce->ce_link != NULL)
	  {
	    /* yes, already had a link. See if specify same link */
	    if ( strcasecmp(p_ce->ce_link,param) != 0 )
	      {
		/*no, they do not specify the same link */
		p_ce->ce_error=ERROR_PARSE_PARAMETER;
		p_ce->ce_string = "ELink";
		return ERROR_PARSE_LINK;
	      }
	    else
	      {
		/* they specify the same link, do nothing*/
		;
	      }
	  }
	else
	  {
	    /* first link specification*/
	    /* assume unescape always results in a smaller string */
	    p_ce->ce_link = param;
	    ap_unescape_url((char *) p_ce->ce_link);
	  }
      }
    else if ( strcasecmp( p, "reset") == 0 )
      {
	re = param;
	if (strcasecmp(re, "access") == 0 )
	  {
	    /* a reset=access command */
	    if (p_ce->ce_reset != reset_notdef)
	      if (p_ce->ce_reset != reset_access)
		{
		  /* duplicate, not equal reset commands */
		  p_ce->ce_error=ERROR_PARSE_PARAMETER;
		  p_ce->ce_string = "EParm";
		  return ERROR_PARSE_PARAMETER;
		}
	    p_ce->ce_reset=reset_access;
	  }
	else if ( strcasecmp( re, "delete") == 0 )
	  {
	    /*  a reset=delete command */
	    if (p_ce->ce_reset != reset_notdef)
	      if (p_ce->ce_reset != reset_delete)
		{
		  /* duplicate, not equal reset commands */
		  p_ce->ce_error=ERROR_PARSE_PARAMETER;
		  p_ce->ce_string = "EParm";
		  return ERROR_PARSE_PARAMETER;
		}
	    p_ce->ce_reset=reset_delete;
	  }
	else if ( strncasecmp( re, "init", 4) == 0 )
	  {
	    ap_unescape_url((char *) re);
	    /*  a reset=init:counting_number command */
	    if (p_ce->ce_reset != reset_notdef)
	      if (p_ce->ce_reset != reset_init)
		{
		  /* duplicate, not equal reset commands */
		  p_ce->ce_error=ERROR_PARSE_PARAMETER;
		  p_ce->ce_string = "EParm";
		  return ERROR_PARSE_PARAMETER;
		}
	    re_init = strchr(re,':');
	    if ( re_init == NULL )
	      {
		p_ce->ce_error=ERROR_PARSE_PARAMETER;
		p_ce->ce_string = "EParm";
		return ERROR_PARSE_PARAMETER;
	      }
	    if (re_init[1] == '\0')
	      {
		p_ce->ce_error=ERROR_PARSE_PARAMETER;
		p_ce->ce_string = "EParm";
		return ERROR_PARSE_PARAMETER;
	      }
	    p_ce->ce_reset=reset_init;
	    p_ce->ce_reset_val = strtol( &re_init[1], (char **)NULL, 10 );
	  }
	else
	  {
	    /* must be a reset=counting_number command */
	    if (p_ce->ce_reset != reset_notdef)
	      if (p_ce->ce_reset != reset_value)
		{
		  /* duplicate, not equal reset commands */
		  p_ce->ce_error=ERROR_PARSE_PARAMETER;
		  p_ce->ce_string = "EParm";
		  return ERROR_PARSE_PARAMETER;
		}
	    p_ce->ce_reset=reset_value;
	    p_ce->ce_reset_val = strtol( re, (char **)NULL, 10 );
	  }
      }
    else if ( strcasecmp( p, "width") == 0 )
      {
      wi = param;
      if (p_ce->ce_width != 0)
	{
	  p_ce->ce_error=ERROR_PARSE_PARAMETER;
	  p_ce->ce_string = "EParm";
	  return ERROR_PARSE_PARAMETER;
	}
      p_ce->ce_width=1;
      if (( p_ce->ce_width_val = strtol( wi, (char **)NULL, 10 )) > MAX_WIDTH_CHAR_DISPLAY )
	{
	p_ce->ce_width = MAX_WIDTH_CHAR_DISPLAY;
	}
      }
    else if ( strcasecmp( p, "font") == 0 )
      {
      fo = param;
      if (p_ce->ce_font != NULL)
	{
	  /* duplicate font= specified */
	  p_ce->ce_error=ERROR_PARSE_PARAMETER;
	  p_ce->ce_string = "EParm";
	  return ERROR_PARSE_PARAMETER;
	}
      if ( strstr( fo, "../" ) == NULL )
	{ 
	p_ce->ce_font = apr_pstrdup( r->pool, fo );
	fstatus=apr_file_open(&fp, apr_pstrcat(r->pool, sza_fontsdir, p_ce->ce_font, ".ttf", NULL),
			      APR_READ | APR_BINARY,
			      APR_OS_DEFAULT, r->pool);
	if ( fstatus != APR_SUCCESS )
	  {
	    p_ce->ce_error=ERROR_PARSE_FONT;
	    p_ce->ce_string = "EFont";
	    return ERROR_PARSE_FONT;
	  }
	else
	  {
	    apr_file_close(fp);
	  }
	}
      }
    else if (( strcasecmp( p, "ignore") == 0 ) && p_ce->ce_ignore == 0 )
      {
      ig = param;
      if ( ( strcmp( ig, p_ce->ce_hostname ) == 0 ) ||
	   ( strcmp( ig, p_ce->ce_ipaddr ) == 0 )      )
	{
	  p_ce->ce_ignore = 1;
	}
      }
    else if ( strcasecmp( p, "bgcolor") == 0 )
      {
	if (p_ce->ce_bgcolor != 0)
	  {
	    p_ce->ce_error=ERROR_PARSE_PARAMETER;
	    p_ce->ce_string = "EParm";
	    return ERROR_PARSE_PARAMETER;
	  }
	if (( bg = param) != NULL )
	  {
	    countm_rgb_color( bg, &p_ce->ce_bgcolor_val );
	    p_ce->ce_bgcolor=1;
	  }
      }
    else if ( strcasecmp( p, "text") == 0 )
      {
	if (p_ce->ce_text != 0)
	  {
	    p_ce->ce_error=ERROR_PARSE_PARAMETER;
	    p_ce->ce_string = "EParm";
	    return ERROR_PARSE_PARAMETER;
	  }
	if (( tx = param) != NULL )
	  {
	    countm_rgb_color( tx, &p_ce->ce_text_val );
	    p_ce->ce_text=1;
	  }
      }
    else if ( strcasecmp( p, "point") == 0 )
      {
	pt = param;
	if (p_ce->ce_point != 0)
	  {
	    p_ce->ce_error=ERROR_PARSE_PARAMETER;
	    p_ce->ce_string = "EParm";
	    return ERROR_PARSE_PARAMETER;
	  }
	p_ce->ce_point_val = strtol( pt, (char **)NULL, 10 );
	p_ce->ce_point=1;
      }
    else if ( strcasecmp( p, "image") == 0 )
      {
	it = param;
	if (p_ce->ce_imgtype != img_notdef)
	  {
	    p_ce->ce_error=ERROR_PARSE_PARAMETER;
	    p_ce->ce_string = "EParm";
	    return ERROR_PARSE_PARAMETER;
	  }
	if ( strcasecmp( it, "jpeg" ) == 0 )
	  {
	    p_ce->ce_imgtype=img_jpeg;
	  }
	else if (strcasecmp( it, "png" ) == 0)
	  {
	    p_ce->ce_imgtype=img_png;
	  }
	else
	  {
	    p_ce->ce_error=ERROR_PARSE_PARAMETER;
	    p_ce->ce_string = "EParm";
	    return ERROR_PARSE_PARAMETER;
	  }
      }
    else if ( strcasecmp( p, "count") == 0)
      {
	ct = param;
	if (p_ce->ce_count != count_notdef)
	  {
	    p_ce->ce_error=ERROR_PARSE_PARAMETER;
	    p_ce->ce_string = "EParm";
	    return ERROR_PARSE_PARAMETER;
	  }
	if ( strcasecmp( ct, "norm_inc") == 0)
	  {
	    p_ce->ce_count=count_norm_inc;
	  }
	else if ( strcasecmp( ct, "norm_dec") == 0)
	  {
	    p_ce->ce_count=count_norm_dec;
	  }
	else if ( strcasecmp( ct, "inc") == 0)
	  {
	    p_ce->ce_count=count_inc;
	  }
	else if ( strcasecmp( ct, "dec") == 0)
	  {
	    p_ce->ce_count=count_dec;
	  }
	else
	  {
	    p_ce->ce_error=ERROR_PARSE_PARAMETER;
	    p_ce->ce_string = "EParm";
	    return ERROR_PARSE_PARAMETER;
	  }
      }
    else if ( (strcasecmp( p, "rand") == 0) || (strcasecmp( p, "random") == 0) )
      {
      ra = param;
      if (p_ce->ce_random != random_notdef)
	{
	  p_ce->ce_error=ERROR_PARSE_PARAMETER;
	  p_ce->ce_string = "EParm";
	  return ERROR_PARSE_PARAMETER;
	}
      if ( strcasecmp( ra, "true" ) == 0 )
	{
	  p_ce->ce_random=random_true;
        }
      else if ( strcasecmp( ra, "true_access" ) == 0 )
	{
	  p_ce->ce_random=random_true_access;
        }
      else if ( strcasecmp( ra, "false" ) == 0 )
	{
	  p_ce->ce_random=random_false;
        }
      else
	{
	  p_ce->ce_error=ERROR_PARSE_PARAMETER;
	  p_ce->ce_string = "EParm";
	  return ERROR_PARSE_PARAMETER;
	}
      }
    else if ( strcasecmp( p, "insert") == 0 )
      {
	ra = param;
	ret = countm_misc_dza_strcat(p_ce->ce_insert_buffer,ra,sizeof(p_ce->ce_insert_buffer));
	if ( (ret == 1) && (p_ce->ce_insert==0) ) p_ce->ce_insert=1;
      }
    } /* while*/
  /* well, we seemed to have parsed the command line, now see if have reasonable values */
  if (p_ce->ce_link == NULL)
    {
      /* all queries must specify a link */
      p_ce->ce_error=ERROR_PARSE_LINK;
      p_ce->ce_string = "ELink";
      return ERROR_PARSE_LINK;
    }  
  return 0;
} /* countm_parse_cmdline */

/**
 *@fn countm_method_handler(request_rec *r)
 *@brief Handler for the method.
 *Processes the request given in r->args into parameters,
 *and completes request.
 *@param r The request record of the method.
 *@return Status value. Possible values are?
 */
static int
countm_method_handler(request_rec *r)
{
  struct countent ce;
  char * cerr=NULL;
  int ret;
  apr_status_t fstatus;

#ifdef DEBUG
  dbg_log = fopen(DEBUG_LOG_FILE_STRING, "w+");
  if (dbg_log==NULL)
    {
      /* You will know it...*/
        ap_log_perror(APLOG_MARK, DEFAULT_LOGLEVEL, 0, r->pool,
		      "Open failed: %s", DEBUG_LOG_FILE_STRING);
    }
  debug_out(dbg_log, "mod_countm: %s\n", version); 
  debug_out(dbg_log,"\n%s: r->unparsed_uri=%s\n", __FUNCTION__, r->unparsed_uri);
  ap_log_perror(APLOG_MARK, DEFAULT_LOGLEVEL, 0, r->pool,
		"URI: %s", r->unparsed_uri);
#endif
  if ( strcasecmp(r->handler,MODULE_HANDLER_ID_STRING)==0 )
    {
      debug_out(dbg_log,"%s:REQUEST: ACCEPTED\n", __FUNCTION__);
    }
else
    {
      debug_out(dbg_log,"%s: REQUEST: DECLINED\n\n", __FUNCTION__);
#ifdef DEBUG
      fclose(dbg_log);
#endif
      return DECLINED;
    }

  if (r->parsed_uri.is_initialized==0)
    {
      fstatus = apr_uri_parse(r->pool, r->unparsed_uri, &r->parsed_uri);
      if (fstatus != APR_SUCCESS)
	{
	  ap_log_perror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, r->pool,
			"Unparsing URI failed: %s", r->unparsed_uri);
	}
    }
  if (strcmp(r->parsed_uri.path,MODULE_URI_PATH_STRING)!=0)
    {
      debug_out(dbg_log,"%s: DECLINED because of path\n\n", __FUNCTION__);
#ifdef DEBUG
      fclose(dbg_log);
#endif
      return DECLINED;
    }

  ret =  countm_parse_cmdline(r, &ce);
  if (ret != 0)
    {
      goto err;
    }

  /* if here, then we must have a valid link, and no major errors (hopefully) */
  /* first, lock database */
  debug_out(dbg_log,"%s: ATTEMPT: lock database %s\n", __FUNCTION__,sza_lockfile);
  ret = countm_lock(sza_lockfile,r->pool);
  if (ret==0)
    {
      debug_out(dbg_log,"%s: FAILED: lock database %s\n", __FUNCTION__,sza_lockfile);
      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, APR_EGENERAL, r,
		    "Lockfile failed: %s", sza_lockfile);
      ce.ce_string = "Edb";
      ce.ce_error=1;
      goto err;
    }
  debug_out(dbg_log,"%s: SUCCESS: Lock database %s\n", __FUNCTION__,sza_lockfile);

  /* open the databases*/
  debug_out(dbg_log,"%s: ATTEMPT: open databases\n", __FUNCTION__);
  ret = countm_db_open_databases(&ce,r);
  if (ret == -1)
    {
      debug_out(dbg_log,"%s: FAILED: open databases\n", __FUNCTION__);
      goto err1;
    }
  debug_out(dbg_log,"%s: SUCCESS: open databases\n", __FUNCTION__);

  /* see if link is in DB_COUNT */
  memset(&ce.ce_dbc_rec,0,sizeof(ce.ce_dbc_rec));
  debug_out(dbg_log,"%s: Is Link in DB_COUNT\n", __FUNCTION__);
  ret = countm_db_count_get_link(&ce);
  if (ret == -1)
    {
      goto err2;
    }
  else if (ret==0)
    {
      /* link is not in DB_COUNT*/
      debug_out(dbg_log,"%s: Link Not in DB_COUNT\n", __FUNCTION__);
    }
  else
    {
      /* link is in DB_COUNT*/
      debug_out(dbg_log,"%s: Link in DB_COUNT\n", __FUNCTION__);
      /* merge info into countent */
      countm_merge_info(&ce, r);
    }

  ret = countm_handler_reset(&ce,r);
  if (ret == 1) goto done;
  ret = countm_handler_random(&ce,r);
  if (ret == 1) goto done;
  countm_handler_updatecounter(&ce,r);
 done:
  countm_setdefaults(r,&ce);
  countm_display_string(&ce, r);

  /* close the database */
  debug_out(dbg_log,"%s: closing databases\n", __FUNCTION__);
  countm_db_close_databases(&ce);

  /* unlock the database */  
  debug_out(dbg_log,"%s: Unlocking lockfile\n", __FUNCTION__);
  ret = countm_unlock(sza_lockfile, r->pool);
  if (ret==0)
    {
      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, APR_EGENERAL, r,
		    "Unlock failed: %s", sza_lockfile);
    }
  debug_out(dbg_log,"%s: Returning OK\n", __FUNCTION__);
#ifdef DEBUG
      fclose(dbg_log);
#endif
    return OK;
 err2:
    /* close the databases*/
  debug_out(dbg_log,"%s: ERROR: closing databases\n", __FUNCTION__);
  countm_db_close_databases(&ce);
 err1:
  /* unlock the database */  
  debug_out(dbg_log,"%s: ERROR: Unlocking database\n", __FUNCTION__);
  if (countm_unlock(sza_lockfile, r->pool)==0)
    {
      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, APR_EGENERAL, r,
		    "Unlock failed: %s", sza_lockfile);
    }
 err:
      /* take the info we have and send back an image */
      countm_setdefaults(r,&ce);
      if ( countm_display_string(&ce, r) == 0 )
	{
	  debug_out(dbg_log, "%s: ERROR: return OK\n\n", __FUNCTION__);
#ifdef DEBUG
	  fclose(dbg_log);
#endif
	  return OK;
	}
      else
	{
	  ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, 0, r,
			"mod_countm: unable to display %s", cerr);
	}
      debug_out(dbg_log, "%s: return(ERROR) HTTP_NOT_FOUND\n\n", __FUNCTION__);
#ifdef DEBUG
      fclose(dbg_log);
#endif
      return HTTP_NOT_FOUND;     
}/* countm_method_handler */

/**
 *@fn countm_set_lockfile(cmd_parms *parms, void *dummy, char *arg)
 *@brief Sets the sza_lockfile string from value given in configuration file.
 *Called during the per-server-configuration if specified in the config file.
 *@param parms The command data structure.
 *@param dummy Unused, since no per-directory-configuration possible.
 *@param arg The value specified in the configuration script.
 *@return NULL
 */
static const char *
countm_set_lockfile(cmd_parms *parms, void *dummy, const char *arg)
{
  sza_lockfile = (char *) arg;
  return NULL;
}

/**
 *@fn countm_set_fontsdir(cmd_parms *parms, void *dummy, char *arg)
 *@brief Sets the sza_fontsdir string from value given in configuration file.
 *Called during the per-server-configuration if specified in the config file.
 *@param parms The command data structure.
 *@param dummy Unused, since no per-directory-configuration possible.
 *@param arg The value specified in the configuration script.
 *@return NULL
 */
static const char *
countm_set_fontsdir(cmd_parms *parms, void *dummy, const char *arg)
{
  sza_fontsdir = (char *) arg;
  return NULL;
}

/**
 *@fn countm_set_font(cmd_parms *parms, void *dummy, char *arg)
 *@brief Sets the sza_font string from value given in configuration file.
 *Called during the per-server-configuration if specified in the config file.
 *@param parms The command data structure.
 *@param dummy Unused, since no per-directory-configuration possible.
 *@param arg The value specified in the configuration script.
 *@return NULL
 */
static const char *
countm_set_font(cmd_parms *parms, void *dummy, const char *arg)
{
  sza_font = (char *) arg;
  return NULL;
}

/**
 *@fn countm_set_db_environment_home(cmd_parms *parms, void *dummy, char *arg)
 *@brief Sets the sza_environment_home string from value given in configuration file.
 *Called during the per-server-configuration if specified in the config file.
 *@param parms The command data structure.
 *@param dummy Unused, since no per-directory-configuration possible.
 *@param arg The value specified in the configuration script.
 *@reutrn NULL
 */
static const char *
countm_set_db_environment_home(cmd_parms *parms, void *dummy, const char *arg)
{
  sza_db_environment_home = (char *) arg;
  return NULL;
}

/**
 *@fn countm_set_db_file(cmd_parms *parms, void *dummy, char *arg)
 *@brief Sets the sza_db_file string from value given in configuration file.
 *Called during the per-server-configuration if specified in the config file.
 *@param parms The command data structure.
 *@param dummy Unused, since no per-directory-configuration possible.
 *@param arg The value specified in the configuration script.
 *@return NULL
 */
static const char *
countm_set_db_file(cmd_parms *parms, void *dummy, const char *arg)
{
  sza_db_file = (char *) arg;
  return NULL;
}

/**
 *@fn countm_set_db_log_file(cmd_parms *parms, void *dummy, char *arg)
 *@brief Sets the sza_db_log_file string from value given in configuration file.
 *Called during the per-server-configuration if specified in the config file.
 *@param parms The command data structure.
 *@param dummy Unused, since no per-directory-configuration possible.
 *@param arg The value specified in the configuration script.
 *@return NULL
 */
static const char *
countm_set_db_log_file(cmd_parms *parms, void *dummy, const char *arg)
{
  sza_db_log_file = (char *) arg;
  return NULL;
}

/**
 *@fn countm_lock(const char * const sza_fn, apr_pool_t * p)
 *@brief Locks the database.
 *Locks the database by creating a lock file.
 *@param sza_fn The name of the lockfile to create.
 *@param p The pool to allocate memory from.
 *@return 1=Success 0=Failure.
 */
static int
countm_lock(const char * const sza_fn, apr_pool_t * p)
{
  int tries=UNLOCK_ATTEMPTS;
  apr_file_t * fp;
  apr_status_t fstatus;

  while (tries>0)
    {
      fstatus=apr_file_open(&fp, sza_fn,
			    APR_READ | APR_WRITE | APR_CREATE | APR_EXCL,
			    APR_OS_DEFAULT, p);
      if (fstatus!=APR_SUCCESS)
	{
	  /* failed, attempt to continue */
	  sleep(1);
	  --tries;
	}
      else
	{
	  /* success */
	  apr_file_close(fp);
	  return 1 ;
	}
    }
  return 0;
}

/**
 *@fn countm_unlock(const char * const sza_fn, apr_pool_t * p)
 *@brief Unlocks the database.
 *Unlocks the database by removing the lock file.
 *@param sza_fn The name of the lockfile to remove.
 *@param p The pool to allocate memory from.
 *@return 1=Success 0=Failure.
 */
static int
countm_unlock(const char * const sza_fn, apr_pool_t * p)
{
  apr_status_t fstatus;

  fstatus=apr_file_remove(sza_fn, p);
  if (fstatus!=APR_SUCCESS)
    {
      return 0;
    }
  else
    {
      return 1;
    }
}

/**
 *@fn countm_rgb_color(const char *hexnum, struct rgb *rgb)
 *@brief Manipulates rgb color values.
 *@param hexnum String representing 6 consecutive hex digits.
 *@param rgb Pointer to allocated return structure.
 *@returns The colors within the rgb structure.
 */
static void
countm_rgb_color(const char *hexnum, struct rgb *rgb)
{
  long hexcolor;
  int mask=0xFF;

  hexcolor=strtol(hexnum, (char **)NULL, 16);
  rgb->rgb_b=(hexcolor & mask);
  rgb->rgb_g=((hexcolor >> 8) & mask);
  rgb->rgb_r=((hexcolor >> 16) & mask);
  return;
}

/**
 *@fn countm_apmod_sink(void *r, char *buf, int len)
 *@brief For gd library
 *@param r request record.
 *@param buf buffer containing the image (sink).
 *@param len Number of bytes to send/ size of sink.
 *@returns The number of bytes sent from the sink.
 */
static int
countm_apmod_sink(void *r, const char *buf, int len)
{
  struct request_rec * rr=(request_rec *) r;
  int bytes_sent;

  /* instead of writing  on stdout, use apache's api */
  bytes_sent = ap_rwrite( buf, len, rr);
  return bytes_sent;
}

/**
 *@fn display_string(struct countent * ce, request_rec *r)
 *@brief displays the string back to the requesting client.
 *@param ce The countent structure with data for request.
 *@param r The request record for this request.
 *@return Status code.
 */
static int
countm_display_string(struct countent * p_ce, request_rec *r)
{
  gdSink ap_sink;
  gdImagePtr im;
  int bgcolor, text, x, y;
  int brect[8];
  apr_status_t fstatus;
  char *err=NULL;
  apr_size_t bytes_sent=0, bytes_to_send;
  FILE * fp_jpeg;
  apr_file_t * fp_ap_jpeg;
  char * fn_jpeg;
  void * p_jpeg_img;
  int jpeg_method;

  /* some final details prior to displaying string */
  p_ce->ce_font = apr_pstrcat(r->pool, sza_fontsdir, p_ce->ce_font, ".ttf", NULL);
  debug_out(dbg_log,"%s: FONT: %s\n", __FUNCTION__, p_ce->ce_font);
  debug_out(dbg_log,"%s: STRING: %s\n", __FUNCTION__, p_ce->ce_string);


  ap_sink.context=(void *) r;
  ap_sink.sink=countm_apmod_sink;

  /* obtain brect to size image */
  err = gdImageStringFT(NULL, &brect[0], 0, p_ce->ce_font,
			(double) p_ce->ce_point_val, 0.0, 0, 0, p_ce->ce_string );
  if (err!=NULL)
    {
      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, 0, r,
		    "gdImageStringFT: %s", err);
      return -1;
    }
  x = brect[2] - brect[6] + 10;
  y = brect[3] - brect[7] + 10;
  im = gdImageCreate(x,y);
  x = 3 - brect[6];
  y = 3 - brect[7];
  /* background color, first allocated */
  bgcolor = gdImageColorResolve(im, p_ce->ce_bgcolor_val.rgb_r,
				p_ce->ce_bgcolor_val.rgb_g, p_ce->ce_bgcolor_val.rgb_b);
  text = gdImageColorResolve(im, p_ce->ce_text_val.rgb_r, p_ce->ce_text_val.rgb_g,
			     p_ce->ce_text_val.rgb_b);
  err = gdImageStringFT(im, &brect[0], text, p_ce->ce_font, (double) p_ce->ce_point_val,
			0.0, x, y, p_ce->ce_string);
  if (err!=NULL)
    {
      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, 0, r,
		    "gdImageStringFT: %s", err);
      gdImageDestroy(im);
      return -1;
    }
  apr_table_set(r->headers_out, "Pragma", "no-cache");
  apr_table_set(r->headers_out, "Cache-Control", "no-cache");
  r->no_cache=1;
#ifdef USE_PNG_IF_ACCEPTABLE
  if ( p_ce->ce_imgtype==img_jpeg )
      {
	accept = apr_table_get( r->headers_in, "Accept" );
	if ( strstr( (char *)accept, "image/png" ))
	  {
	    p_ce->ce_imgtype=img_png;
	  }
      }
#endif
  if (p_ce->ce_imgtype==img_png)
    {
      ap_set_content_type(r,"image/png");
      gdImagePngToSink( im, &ap_sink );
      gdImageDestroy( im );
    }
  else if (p_ce->ce_imgtype==img_jpeg)
    {
      ap_set_content_type(r,"image/jpeg");
      fn_jpeg = apr_psprintf(r->pool, "%s%scountm%i.jpg", TEMPORARY_WRITEABLE_DIRECTORY_PATH,"/",(int)getpid());
      fp_jpeg = fopen(fn_jpeg, "w+b");
      if (fp_jpeg==NULL)
	{
	  ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, 0, r,
			"fopen failed: %s", fn_jpeg);
	  gdImageDestroy( im );
	  return -1;
	}
      gdImageJpeg( im, fp_jpeg, 95 );
      bytes_to_send = ftell(fp_jpeg);
      jpeg_method=0;
      if (jpeg_method==0)
	{
	  p_jpeg_img = (void *) apr_palloc(r->pool, bytes_to_send);
	  if (p_jpeg_img == NULL)
	    {
	      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, APR_EGENERAL, r,
			    "apr_palloc failed");
	      return -1;
	    }
	  fseek(fp_jpeg, 0, SEEK_SET);
	  bytes_sent = fread(p_jpeg_img, sizeof(char), bytes_to_send, fp_jpeg);
	  if (bytes_sent != bytes_to_send)
	    {
	      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, APR_EGENERAL, r,
			    "Reading jpeg file failed: %d, %d",
			    bytes_to_send, bytes_sent);
	      return -1;
	    }
	  fclose(fp_jpeg);
#ifndef DEBUG
	  apr_file_remove(fn_jpeg,r->pool);
#endif
	  bytes_sent = ap_rwrite( p_jpeg_img, bytes_to_send, r);
	  if (bytes_sent != bytes_to_send)
	    {
	      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, APR_EGENERAL, r,
			    "ap_rwrite failed: %s", fn_jpeg);
	      return -1;
	    }
	  gdImageDestroy( im );
	}
      else
	{
	  fclose(fp_jpeg);
	  apr_file_perms_set(fn_jpeg, APR_OS_DEFAULT);
	  fstatus = apr_file_open(&fp_ap_jpeg, fn_jpeg,
				  APR_READ | APR_WRITE | APR_BINARY, /* | APR_DELONCLOSE,*/
				  APR_OS_DEFAULT, r->pool);
	  if (fstatus != APR_SUCCESS)
	    {
	      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, r,
			    "ReOpen failed: %s", fn_jpeg);
	      return -1;
	    }
	  bytes_sent = 0;
	  ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, r,
			"About to send: %s, %d", fn_jpeg, bytes_to_send);
	  fstatus=ap_send_fd( fp_ap_jpeg, r, 0, bytes_to_send, &bytes_sent);
	  if (fstatus != APR_SUCCESS)
	    {
	      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, r,
			    "Sending jpeg image %s", fn_jpeg);
	      apr_file_close(fp_ap_jpeg);
	      gdImageDestroy( im );
	      return -1;
	    }
	  ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, fstatus, r,
			"Sent: %s, %d, %d",
			fn_jpeg, bytes_to_send, bytes_sent);
	  apr_file_close( fp_ap_jpeg );
	  gdImageDestroy( im );
	}
    }
  else
    return -1;
  return 0;
} /* display_string */

/**
 * @fn countm_misc_dza_strcat( char * const buf, char const * const szstr, const int bsize)
 * @brief countm miscellaneous double zero ascii string concat function
 * appends a zero terminated ascii string to a double zero terminated list of zero terminated strings.
 * A nice double zero terminated string library would be nice.
 * @param buf The buffer to append
 * @param bsize The buffer size, in bytes.
 * @return 0=failed, buffer too small, invalid szstr, 1=success
 **/
static int countm_misc_dza_strcat( char * const buf, char const * const szstr, const int bsize)
{
  char * pc=buf;
  int j=0;
  int prev=0;
  int len;

  if (bsize <= 0) return 0;
  if (szstr[0]=='\0') return 0;

  len = strlen(szstr);
  if ( (len + 1) >= bsize) return 0;
  if ( (buf[0]=='\0') && (buf[1]=='\0') )
    {
      strcpy(buf,szstr);
      buf[len+1]='\0';
      return 1;
    }
  while ( j < bsize )
    {
      if ( (prev == 1) && (*pc == '\0') ) break;
      if ( *pc == '\0' )
	prev=1;
      else
	prev=0;
      ++j;++pc;
    }
  printf("%s,%d\n",pc,j);

  if (j >= bsize )
    {
      return 0;
    }
  if ( (abs(pc - buf) + len + 2) > bsize ) /* +2 => two '\0'*/
    return 0;
  memcpy(pc,szstr,len+1);
  pc[len+1]=0;
  return 1;
}/* countm_misc_dza_strcat */

#ifdef DEBUG
void
countm_debug_print_countent(FILE * fp, struct countent * p_ce)
{
  if (p_ce == NULL )
    {
      debug_out(dbg_log, "%s: p_ce is NULL\n", __FUNCTION__);
      return;
    }
  fprintf(fp,"ce_link=%s\n",p_ce->ce_link);
  fprintf(fp,"ce_reset=%i\n",(int)p_ce->ce_reset);
  fprintf(fp,"ce_reset_val=%i\n",(int)p_ce->ce_reset_val);
  fprintf(fp,"ce_width=%i\n",p_ce->ce_width);
  fprintf(fp,"ce_width_val=%i\n",(int)p_ce->ce_width_val);
  fprintf(fp,"ce_font=%s\n",p_ce->ce_font);
  fprintf(fp,"ce_ignore=%i\n",p_ce->ce_ignore);
  fprintf(fp,"ce_bgcolor=%i\n",p_ce->ce_bgcolor);
  fprintf(fp,"ce_bgcolor_val=%i,%i,%i\n",p_ce->ce_bgcolor_val.rgb_r,p_ce->ce_bgcolor_val.rgb_g,p_ce->ce_bgcolor_val.rgb_b);
  fprintf(fp,"ce_text=%i\n",p_ce->ce_text);
  fprintf(fp,"ce_text_val=%i,%i,%i\n",p_ce->ce_text_val.rgb_r,p_ce->ce_text_val.rgb_g,p_ce->ce_text_val.rgb_b);
  fprintf(fp,"ce_point=%i\n",p_ce->ce_point);
  fprintf(fp,"ce_point_val=%i\n",p_ce->ce_point_val);
  fprintf(fp,"ce_imgtype=%i\n",p_ce->ce_imgtype);
  fprintf(fp,"ce_count=%i\n",p_ce->ce_count);
  fprintf(fp,"ce_random=%i\n",(int) p_ce->ce_random);
  fprintf(fp,"ce_insert=%i\n",p_ce->ce_insert);
  fprintf(fp,"ce_error=%i\n",(int)p_ce->ce_error);
  fprintf(fp,"ce_hostname=%s\n",p_ce->ce_hostname);
  fprintf(fp,"ce_ipaddr=%s\n",p_ce->ce_ipaddr);
  fprintf(fp,"ce_hostname_is_ipaddr=%i\n",p_ce->ce_hostname_is_ipaddr);
  fprintf(fp,"ce_string=%s\n",p_ce->ce_string);
  return;
} /* countm_debug_print_countent */

void
countm_debug_print_db_count_record(FILE * fp, struct db_count_record * p)
{
  if ( p == NULL )
    {
      debug_out(dbg_log, "%s: p is NULL\n", __FUNCTION__);
      return;
    }

  fprintf(fp,"cnt=%u\n",(unsigned int)p->cnt);
  fprintf(fp,"font=%s\n",p->font);
  fprintf(fp,"imgtype=%i\n",(int)p->imgtype);
  fprintf(fp,"count=%i\n",(int)p->count);
  fprintf(fp,"random=%i\n",(int)p->random);
  fprintf(fp,"status=0x%X\n",p->status);
  fprintf(fp,"width_val=%i\n",p->width_val);
  fprintf(fp,"point_val=%i\n",p->point_val);
  fprintf(fp,"bgcolor_val %i, %i, %i\n",p->bgcolor_val.rgb_r, p->bgcolor_val.rgb_g, p->bgcolor_val.rgb_b);
  fprintf(fp,"text_val %i, %i, %i\n",p->text_val.rgb_r, p->text_val.rgb_g, p->text_val.rgb_b);
  return;
} /* countm_debug_print_db_count_record */

/**
 *countm_debug_print_request_rec
 */
void
countm_debug_print_request_rec(FILE * fp, request_rec * r)
{
  uint64_t u64_i=(uint64_t)r->allowed;

  fprintf(fp,"allowed=0x%016X\n", u64_i);
  fprintf(fp,"allowed=0x%016X\n", u64_i);
  return;
} /* countm_debug_print_request_rec */
#endif

/**
 * @fn countm_db_open_databases(struct countent * p_ce, request_rec * r)
 * @brief opens the db4 databases environment, DB_ACCESS, DB_COUNT
 * Creates the db4 environment, opens both DB_ACCESS and DB_COUNT
 * @param p_ce the request data
 * @param r the api request data
 * @return -1=ERROR, 1=success
 */
static int
countm_db_open_databases(struct countent * p_ce, request_rec * r)
{

  /* open error log */
  p_ce->ce_fp_log = fopen(sza_db_log_file, "a+");
  if (p_ce->ce_fp_log==NULL)
    {
      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, APR_EGENERAL, r,
		    "Open failed: %s", sza_db_log_file);
      p_ce->ce_string="EFile";
      p_ce->ce_error=1;
      return -1;
    }
  p_ce->ce_db_log_file=1;

  p_ce->ce_cachesize=0;
  p_ce->ce_dbenv = countm_db_env_init(sza_db_environment_home, sza_db_error_log_prefix,
		      p_ce->ce_cachesize, p_ce->ce_fp_log);
  if (p_ce->ce_dbenv==0)
    {
      ap_log_rerror(APLOG_MARK, DEFAULT_LOGLEVEL, APR_EGENERAL, r,
		    "DB_ENV create failed: %s", sza_db_file);
      fprintf(p_ce->ce_fp_log,"ERROR: %s: DB_ENV create failed\n",__FUNCTION__);
      p_ce->ce_string = "EFile";
      goto err;
    }

  p_ce->ce_dbp_a = countm_db_open(p_ce->ce_dbenv, sza_db_file, DB_ACCESS, DB_HASH, DB_CREATE,
		  DB_DUP | DB_DUPSORT, 0664, 0);
  if (p_ce->ce_dbp_a==NULL)
    {
      fprintf(p_ce->ce_fp_log, "ERROR: %s: db_open: access\n",__FUNCTION__);
      p_ce->ce_string = "Edb";
      goto err1;
    }

  p_ce->ce_dbp_c = countm_db_open(p_ce->ce_dbenv, sza_db_file, DB_COUNTER, DB_HASH, DB_CREATE, 0,
		  0664, 0);
  if (p_ce->ce_dbp_c==NULL)
    {
      fprintf(p_ce->ce_fp_log, "ERROR: %s: db_open: counter\n",__FUNCTION__);
      p_ce->ce_string = "Edb";
      goto err2;
    }
  return 1;
  goto err3;
 err3:
  p_ce->ce_dbp_c->close(p_ce->ce_dbp_c,0);
 err2:
  p_ce->ce_dbp_a->close(p_ce->ce_dbp_a,0);
 err1:
  p_ce->ce_dbenv->close(p_ce->ce_dbenv, 0);
 err:
  p_ce->ce_error=1;
  fclose(p_ce->ce_fp_log);
  p_ce->ce_db_log_file=0;
  return -1;

} /* countm_db_open_databases */

/**
 *@fn countm_db_open(DB_ENV * dbenv, const char * fname, const char * dname,
	DBTYPE dtype, u_int32_t dflags_open, u_int32_t dflags_db,
	int dmode, u_int32_t dpgsize)
 *@breif Actually opens a database
 *@param dbenv The environment to open in (returned from countm_db_env_init)
 *@param dname The internal database name.
 *@param dtype The data base type.
 *@param dflags_open The open database flags.
 *@param dflags_db The database flags (DB->set_flags).
 *@param dmode The file creation mode (0664 normal)
 *@param dpgsize The page size for the database (DB->set_page_size)
 *@return NULL=Error, dbp=Success.
 */
static DB *
countm_db_open(DB_ENV * dbenv, const char * fname, const char * dname,
	DBTYPE dtype, u_int32_t dflags_open, u_int32_t dflags_db,
	int dmode, u_int32_t dpgsize)
{
  int ret;
  DB * dbp;

  if ( (dbenv==NULL) || (fname==NULL) ) return NULL;
  if ((ret = db_create(&dbp, dbenv, 0)) != 0)
    {
      dbenv->err(dbenv, ret, "db_create"); 
      return NULL;
    }

  if ((ret = dbp->set_flags(dbp, dflags_db)) != 0)
    {
      dbp->err(dbp, ret, "DB->set_flags");
      goto err;
    }
  if (dpgsize!=0)
    if ((ret = dbp->set_pagesize(dbp, dpgsize)) !=0)
      {
	dbp->err(dbp, ret, "DB->set_pagesize");
	goto err;
      }

#if (DB_VERSION_MAJOR >= 4) && (DB_VERSION_MINOR > 0)
  if ((ret = dbp->open(dbp, (DB_TXN *) NULL, fname, dname, dtype,
		       dflags_open, dmode)) != 0)
#else
  if ((ret = dbp->open(dbp, fname, dname, dtype, dflags_open, dmode)) != 0)
#endif
    {
      dbp->err(dbp, ret, "DB->open: %s:%s", fname, dname);
      goto err;
    }
  return dbp;
 err:
  dbp->close(dbp, 0);
  return (NULL);
}

/**
 *@fn countm_db_close_databases(struct countent * p_ce)
 * closes the db4 database, environment, DB_COUNT, DB_ACCESS
 *@param p_ce request data
 *@return None
*/
static void
countm_db_close_databases(struct countent * p_ce)
{
  p_ce->ce_dbp_a->close(p_ce->ce_dbp_a,0);
  debug_out(dbg_log, "%s: access closed.\n", __FUNCTION__);
  p_ce->ce_dbp_c->close(p_ce->ce_dbp_c,0);
  debug_out(dbg_log, "%s: count.\n", __FUNCTION__);
  p_ce->ce_dbenv->close(p_ce->ce_dbenv, 0);
  debug_out(dbg_log, "%s: dbenv closed.\n", __FUNCTION__);
  fclose(p_ce->ce_fp_log);
} /* countm_db_close_databases */

/**
 *@fn countm_db_env_init(const char * home, const char * prefix, int cachesize,
	    FILE * fp_log)
 *@brief Initialize the db4 database environment.
 *@param home The environment home.
 *@param prefix The prefix for the error log.
 *@param cachesize The cachesize for the database.
 *@param fp_log The FILE * to the error log for the database.
 *@return NULL=error, dbenv if successful
 */
static DB_ENV *
countm_db_env_init(const char * home, const char * prefix, int cachesize,
	    FILE * fp_log)
{
	DB_ENV *dbenv;
	int ret;

	if ((ret = db_env_create(&dbenv, 0)) != 0)
	  {
	    if (fp_log!=NULL) fprintf(fp_log, "%s: db_init: %s\n",
				      prefix, db_strerror(ret));
	    return NULL;
	  }
	if (fp_log!=NULL) dbenv->set_errfile(dbenv, fp_log);
	if (prefix !=NULL) dbenv->set_errpfx(dbenv, prefix);
	if (cachesize !=0) (void)dbenv->set_cachesize(dbenv, 0,
						      (u_int32_t)cachesize, 0);

	if ((ret = dbenv->open(dbenv, home,
			       DB_CREATE | DB_INIT_MPOOL, 0)) != 0)
	  {
	    dbenv->err(dbenv, ret, "DB_ENV->open: %s", home);
	    (void)dbenv->close(dbenv, 0);
	    return NULL;
	  }
	return dbenv;
} /* countm_db_env_init */

/**
 *@fn countm_db_count_get_link(struct countent * p_ce)
 *@brief Attempts to retrieve link from DB_COUNT
 *@param p_ce request data
 * @return 0 = SUCCESS, p_ce->dbc_rec contains the found record
 * 1 = FAILURE link not found in database
 * -1 = ERROR. Some sort of database error
 */
static int
countm_db_count_get_link(struct countent * p_ce)
{
  DBT key;
  DBT data;
  int ret;
  struct db_count_record buf;

  memset( &key, 0, sizeof( key ));
  memset( &data, 0, sizeof( data ));

  key.data = (void *) p_ce->ce_link;
  key.size = strlen( key.data ) + 1;
  /* get count for link from DB_COUNT */
  data.data=(void *) &buf;
  data.size = sizeof(buf);
  debug_out(dbg_log, "%s: ATTEMPT: DB_COUNT: Find key=%s\n", __FUNCTION__,(char *) key.data);
  ret = p_ce->ce_dbp_c->get(p_ce->ce_dbp_c, NULL, &key, &data, 0);
  memcpy(&p_ce->ce_dbc_rec,data.data,sizeof(p_ce->ce_dbc_rec));
  if (ret==DB_NOTFOUND)
    {
      /* link not found */
      debug_out(dbg_log, "%s: RETURNED: DB_COUNT: DB_NOTFOUND\n", __FUNCTION__);
      p_ce->ce_link_in_db_count=0;
      ret=0;
    }
  else if (ret ==0)
    {
      /* value is valid */
      debug_out(dbg_log, "%s: RETURNED: DB_COUNT: FOUND\n", __FUNCTION__);
      p_ce->ce_link_in_db_count=1;
      ret=1;
    }
  else
    {
      /* db error */
      debug_out(dbg_log, "%s: RETURNED: DB_COUNT: ERROR\n", __FUNCTION__);
      p_ce->ce_string = "Edb";
      p_ce->ce_error=1;
      ret = -1;
    }
  return ret;
} /* countm_db_count_get_link */

/**
 *@fn countm_merge_info(struct countent * p_ce, request_rec * r)
 *@brief merges the info from DB_COUNT into query
 * Merges the info from DB_COUNT into the query.
 * Two possiblities: Info from dbase takes precedence, or info from command line takes precedence.
 *@param p_ce request data
 *@param r api request data
 *@return Always returns 0
 */
static int
countm_merge_info(struct countent * p_ce, request_rec * r)
{
  int width, point, bgcolor, text;

  width = ( p_ce->ce_dbc_rec.status & 1);
  point = (p_ce->ce_dbc_rec.status & (1 << 1));
  bgcolor = (p_ce->ce_dbc_rec.status & (1 << 2));
  text = (p_ce->ce_dbc_rec.status & (1 << 3));
  switch (COUNTM_INFORMATION_PRIORITY)
    {
    case COUNTM_INFORMATION_PRIORITY_DATABASE:
      /* database priority */
      debug_out(dbg_log,"%s: MERGE: Priority database\n", __FUNCTION__);
      if (p_ce->ce_dbc_rec.font[0] != '\0')
	{
	  p_ce->ce_font=apr_pstrdup(r->pool,p_ce->ce_dbc_rec.font);
	}
      if (p_ce->ce_dbc_rec.imgtype != img_notdef)
	{
	  p_ce->ce_imgtype = p_ce->ce_dbc_rec.imgtype;
	}
      if (p_ce->ce_dbc_rec.count != count_notdef)
	{
	  p_ce->ce_count = p_ce->ce_dbc_rec.count;
	}
      if (p_ce->ce_dbc_rec.random != random_notdef)
	{
	  p_ce->ce_random = p_ce->ce_dbc_rec.random;
	}
      if (width)
	{
	  p_ce->ce_width_val = p_ce->ce_dbc_rec.width_val;
	  p_ce->ce_width=1;
	}
      if (point)
	{
	  p_ce->ce_point_val = p_ce->ce_dbc_rec.point_val;
	  p_ce->ce_point=1;
	}
      if (bgcolor)
	{
	  memcpy(&p_ce->ce_bgcolor_val, &p_ce->ce_dbc_rec.bgcolor_val, sizeof(p_ce->ce_bgcolor_val));
	  p_ce->ce_bgcolor=1;
	}
      if (text)
	{
	  memcpy(&p_ce->ce_text_val, &p_ce->ce_dbc_rec.text_val, sizeof(p_ce->ce_text_val));
	  p_ce->ce_text=1;
        }
      break;
    case COUNTM_INFORMATION_PRIORITY_COMMANDLINE:
      /* commandline priority */
      debug_out(dbg_log,"%s: MERGE: Priority commandline\n", __FUNCTION__);
      if ( (p_ce->ce_dbc_rec.font[0] != '\0') && (p_ce->ce_font==NULL) )
	{
	  p_ce->ce_font=apr_pstrdup(r->pool,p_ce->ce_dbc_rec.font);
	}
      if ( (p_ce->ce_dbc_rec.imgtype != img_notdef) && (p_ce->ce_imgtype == img_notdef) )
	{
	  p_ce->ce_imgtype = p_ce->ce_dbc_rec.imgtype;
	}
      if ( (p_ce->ce_dbc_rec.count != count_notdef) && (p_ce->ce_count == count_notdef) )
	{
	  p_ce->ce_count = p_ce->ce_dbc_rec.count;
	}
      if ( (p_ce->ce_dbc_rec.random != random_notdef) && (p_ce->ce_random == random_notdef) )
	{
	  p_ce->ce_random = p_ce->ce_dbc_rec.random;
	}
      if (width && (p_ce->ce_width==0) )
	{
	  p_ce->ce_width_val = p_ce->ce_dbc_rec.width_val;
	  p_ce->ce_width=1;
	}
      if (point && ( p_ce->ce_point==0) )
	{
	  p_ce->ce_point_val = p_ce->ce_dbc_rec.point_val;
	  p_ce->ce_point=1;
	}
      if (bgcolor && (p_ce->ce_bgcolor==0) )
	{
	  memcpy(&p_ce->ce_bgcolor_val, &p_ce->ce_dbc_rec.bgcolor_val, sizeof(p_ce->ce_bgcolor_val));
	  p_ce->ce_bgcolor=1;
	}
      if (text && (p_ce->ce_text==0) )
	{
	  memcpy(&p_ce->ce_text_val, &p_ce->ce_dbc_rec.text_val, sizeof(p_ce->ce_text_val));
	  p_ce->ce_text=1;
	}
      break;
    default:
      debug_out(dbg_log,"%s: MERGE: ERROR\n", __FUNCTION__);
      break;
    } /*switch */
  p_ce->ce_db_val = p_ce->ce_dbc_rec.cnt;
  return 0;
} /* countm_merge_info */

/**
 *@fn countm_handler_reset(struct countent * p_ce, request_rec *r)
 *@brief Handles the reset commands
 *Handles the reset commmand.
 *@param p_ce request data
 *@param r api request data
 *@return 1 = true = was a reset command, and handled
 * 0 = false = not handled.
 * -1 = ERROR.
 */
static int
countm_handler_reset(struct countent * p_ce, request_rec *r)
{
  DBT key, data;
  char * pc;
  int len;
  int ret;

  if (p_ce->ce_reset == reset_value)
    {
      /* ce_reset == reset_value */
      /* is link already in DB_COUNT? */
      if (p_ce->ce_link_in_db_count != 1)
	{
	  /* no. link is not in DB_COUNT */
	  /* return success, but indicate "Error" */
	  debug_out(dbg_log, "%s: Not in DB_COUNT\n",__FUNCTION__);
	  p_ce->ce_string = "NoLink";
	  p_ce->ce_error=0;
	  return 1;      
	}
      else
	{
	  /* link is already in DB_COUNT */
	  /* ce_dbc_rec contains old count, ce_reset_val contains new value */
	  /* put new count into DB_COUNT, and return value that was in DB_COUNT */
	  /* do not adjust any other values in database */
	  debug_out(dbg_log, "%s: Already in DB_COUNT\n",__FUNCTION__);
	  p_ce->ce_db_val = p_ce->ce_dbc_rec.cnt;
	  p_ce->ce_dbc_rec.cnt = p_ce->ce_reset_val;
	  memset(&key,0,sizeof(key));
	  memset(&data,0,sizeof(data));
	  key.data = (void *) p_ce->ce_link;
	  key.size = strlen(key.data) +1;
	  data.data = (void *) &p_ce->ce_dbc_rec;
	  data.size = sizeof(p_ce->ce_dbc_rec);
	  debug_out(dbg_log, "%s: DB_COUNT: PUT: link=%s: dbc_rec size=%i\n",__FUNCTION__, (char *) key.data, data.size);
	  ret = p_ce->ce_dbp_c->put(p_ce->ce_dbp_c, NULL, &key, &data, 0); /* replace existing key */
	  if (ret != 0)
	    {
	      /* some sort of database error */
	      debug_out(dbg_log, "%s: DB_COUNT: PUT: ERROR - returned %i\n",__FUNCTION__, ret);
	      p_ce->ce_error=1;
	      p_ce->ce_string="Edb";
	      return 1;
	    }
	  debug_out(dbg_log, "%s: DB_COUNT: PUT: success\n",__FUNCTION__);
	  /* convert count to string */
	  p_ce->ce_string = apr_ltoa(r->pool, p_ce->ce_db_val);
	  p_ce->ce_error=0;
	  return 1;
	}
    }
  else if (p_ce->ce_reset==reset_delete)
    {
      /* if link is in DB_COUNT, delete it*/
      /* if link is not in DB_COUNT, return NoLink error*/
      if (p_ce->ce_link_in_db_count==0)
	{
	  /* link is not in DB_COUNT */
	  p_ce->ce_string="Nolink";
	  p_ce->ce_error=0;
	  return 1;
	}
      else
	{
	  /* delete link from DB_COUNT */
	  memset(&key,0,sizeof(key));
	  key.data = (void *) p_ce->ce_link;
	  key.size = strlen(key.data) +1;
	  debug_out(dbg_log,"%s: RESET: DELETE attempt link=%s\n",__FUNCTION__, (char *) key.data);
	  ret = p_ce->ce_dbp_c->del(p_ce->ce_dbp_c, NULL, &key, 0);
	  if (ret ==  DB_NOTFOUND)
	    {
	      debug_out(dbg_log,"%s: RESET: DELETE: DB_NOTFOUND\n", __FUNCTION__);
	      p_ce->ce_string="NoLink";
	      p_ce->ce_error=0;
	      return 1;
	    }
	  else if ( ret != 0)
	    {
	      debug_out(dbg_log,"%s: RESET: DELETE: ERROR\n", __FUNCTION__);
	      p_ce->ce_string="Edb";
	      p_ce->ce_error=1;
	      return 1;
	    }
	  else
	    {
	      debug_out(dbg_log,"%s: RESET: DELETE: SUCCESS\n", __FUNCTION__);
	      p_ce->ce_string="Cdel";
	      p_ce->ce_error=0;
	      return 1;
	    }
	}
    }
  else if (p_ce->ce_reset==reset_access)
    {
      /* delete access list from DB_ACCESS*/
      memset(&key,0,sizeof(key));
      key.data = (void *) p_ce->ce_link;
      key.size = strlen(key.data) +1;
      debug_out(dbg_log,"%s: RESET: ACCESS attempt link=%s\n",__FUNCTION__, (char *) key.data);
      ret = p_ce->ce_dbp_a->del(p_ce->ce_dbp_a, NULL, &key, 0);
      if (ret == DB_NOTFOUND)
	{
	  debug_out(dbg_log,"%s: RESET: ACCESS: DB_NOTFOUND\n",__FUNCTION__);
	  p_ce->ce_string="NoLink";
	  p_ce->ce_error=0;
	  return 1;
	}
      else if (ret == 0)
	{
	  /* A okay */
	  debug_out(dbg_log,"%s: RESET: DELETE: SUCCESS\n", __FUNCTION__);
	  p_ce->ce_string="Arst";
	  p_ce->ce_error=0;
	  return 1;
	}
      else
	{
	  /* db error */
	  debug_out(dbg_log,"%s: RESET: DELETE: ERROR\n", __FUNCTION__);
	  p_ce->ce_string="Edb";
	  p_ce->ce_error=1;
	  return 1;
	}
    }
  else if (p_ce->ce_reset==reset_init)
    {
      /* ce_reset == reset_init */
      /* is a value already in link? */
      if (p_ce->ce_link_in_db_count != 1)
	{
	  /* no, insert the new count, along with any given information, and return Cadd */
	  debug_out(dbg_log, "%s: Not in DB_COUNT\n",__FUNCTION__);
	  countm_fill_db_count_rec(p_ce);
	  /* put it into DB_COUNT */
	  memset(&key,0,sizeof(key));
	  key.data = (void *) p_ce->ce_link;
	  key.size = strlen(key.data) +1;
	  memset(&data,0,sizeof(data));
	  data.data = (void *) &p_ce->ce_dbc_rec;
	  data.size = sizeof(p_ce->ce_dbc_rec);
	  debug_out(dbg_log, "%s: DB_COUNT: PUT: link=%s: dbc_rec size=%i\n",__FUNCTION__, (char *) key.data, data.size);
	  ret = p_ce->ce_dbp_c->put(p_ce->ce_dbp_c, NULL, &key, &data, 0); /* overwrite existing */
	  if (ret != 0)
	    {
	      /* some sort of database error */
	      debug_out(dbg_log, "%s: DB_COUNT: PUT: ERROR - failed\n",__FUNCTION__);
	      p_ce->ce_error=1;
	      p_ce->ce_string="Edb";
	      return 1;
	    }
	  debug_out(dbg_log, "%s: DB_COUNT: PUT: success\n",__FUNCTION__);
	  /* may be some inserts into DB_ACCESS */
	  if (p_ce->ce_insert == 1)
	    {
	      /* have some hostnames (at least one) to insert into DB_ACCESS */
	      pc = p_ce->ce_insert_buffer;
	      do
		{
		  len = strlen(pc) + 1;
		  memset(&key,0,sizeof(key));
		  key.data = (void *) p_ce->ce_link;
		  key.size = strlen(key.data) +1;
		  memset(&data,0,sizeof(data));
		  data.data = (void *) pc;
		  data.size = len;
		  debug_out(dbg_log, "%s: DB_ACCESS: PUT: link=%s data=%s\n", __FUNCTION__, (char *)key.data, (char *)data.data);
		  ret = p_ce->ce_dbp_a->put(p_ce->ce_dbp_a, NULL, &key, &data, DB_NODUPDATA);
		  if ( (ret == DB_KEYEXIST) || (ret == 0) )
		    {
		      /* success */
		      if (ret == 0)
			{
			  debug_out(dbg_log, "%s: DB_ACCESS: PUT: success\n", __FUNCTION__);
			}
		      else
			{
			  debug_out(dbg_log, "%s: DB_ACCESS: PUT: Already present\n", __FUNCTION__);
			}
		    }
		  else
		    {
		      /* some sort of database error*/
		      debug_out(dbg_log, "%s: DB_ACCESS: PUT: ERROR - failure\n", __FUNCTION__);
		      p_ce->ce_error=1;
		      p_ce->ce_string="Edb";
		      return 1;
		    }
		  pc+=len;
		} while ( *pc != '\0');
	    }
	  p_ce->ce_string = "Cadd";
	  p_ce->ce_error=0;
	  return 1;      
	}
      else
	{
	  /* link is already in DB_COUNT */
	  /* return success, but indicate error */
	  debug_out(dbg_log, "%s: Already in DB_COUNT\n",__FUNCTION__);
	  p_ce->ce_string = "Elink";
	  p_ce->ce_error=0;
	  return 1;
	}
    }
  else
    {
      /* p_ce->ce_reset == reset_notdef)*/ 
      debug_out(dbg_log,"%s: RESET: NOTDEFINED\n", __FUNCTION__);
      return 0;
    }
  return -1;
}/* countm_handler_reset */

/**
 *@fn countm_fill_db_count_rec(struct countent * p_ce)
 *@brief place information into ce_dbc_rec from ce
 *@return Always 0
 */
static int
countm_fill_db_count_rec(struct countent * p_ce)
{

  memset(&p_ce->ce_dbc_rec, 0, sizeof(p_ce->ce_dbc_rec)); /* MAX_FONT_NAME_SIZE */
  p_ce->ce_dbc_rec.imgtype=img_notdef;
  p_ce->ce_dbc_rec.count = count_notdef;
  p_ce->ce_dbc_rec.random = random_notdef;
  if (p_ce->ce_font == NULL)
    {
      ;
    }
  else
    {
      if (strlen(p_ce->ce_font) < MAX_FONT_NAME_SIZE) strcpy(p_ce->ce_dbc_rec.font, p_ce->ce_font);
    }
  if (p_ce->ce_imgtype != img_notdef)
    {
      p_ce->ce_dbc_rec.imgtype = p_ce->ce_imgtype;
    }
  if (p_ce->ce_count != count_notdef)
    {
      p_ce->ce_dbc_rec.count = p_ce->ce_count;
    }
  if (p_ce->ce_random != random_notdef)
    {
      p_ce->ce_dbc_rec.random = p_ce->ce_random;
    }
  if (p_ce->ce_bgcolor == 1)
    {
      memcpy(&p_ce->ce_dbc_rec.bgcolor_val,&p_ce->ce_bgcolor_val, sizeof(p_ce->ce_dbc_rec.bgcolor_val));
      p_ce->ce_dbc_rec.status |= (1 << 2);
    }
  if (p_ce->ce_text == 1)
    {
      memcpy(&p_ce->ce_dbc_rec.text_val,&p_ce->ce_text_val, sizeof(p_ce->ce_dbc_rec.text_val));
      p_ce->ce_dbc_rec.status |= (1 << 2);
    }
  if (p_ce->ce_point == 1)
    {
      p_ce->ce_dbc_rec.point_val = p_ce->ce_point_val;
      p_ce->ce_dbc_rec.status |= (1 << 1);

    }
  if (p_ce->ce_width == 1)
    {
      p_ce->ce_dbc_rec.width_val = p_ce->ce_width_val;
      p_ce->ce_dbc_rec.status |= 1;
    }
  p_ce->ce_dbc_rec.cnt = p_ce->ce_reset_val;
  return 0;
} /* countm_fill_db_count_rec */

/**
 *@fn countm_handler_random(struct countent * p_ce, request_rec *r)
 *@brief  handle the random= cases
 *@param p_ce rquest data
 *@param r api request data
 *@return 1=true=was random, and handled
 *0=not handled, must continue to process
 *-1=some sort of error
 */
static int
countm_handler_random(struct countent * p_ce, request_rec *r)
{
  DBT key, data;
  int ret;

  if (p_ce->ce_random == random_false)
    {
      /* do not handle it */
      return 0;
    }
  else  if (p_ce->ce_random == random_notdef)
    {
      /* should never happen !! right*/
      p_ce->ce_random = random_false;
      return 0;
    }
  /*  if ( (p_ce->ce_random == random_true) || (p_ce->ce_random = random_true_access) )*/
  /* return a random value */
  /* possibly inserting host into DB_ACCESS */
  if (p_ce->ce_random == random_true_access)
    {
      /* insert host into DB_ACCESS */
      memset(&key,0,sizeof(key));
      key.data = (void *) p_ce->ce_link;
      key.size = strlen(key.data) +1;
      memset(&data,0,sizeof(data));
      data.data = (void *) p_ce->ce_hostname;
      data.size = strlen(p_ce->ce_hostname) +1;
      debug_out(dbg_log, "%s: DB_ACCESS: PUT: link=%s data=%s\n", __FUNCTION__, (char *)key.data, (char *)data.data);
      ret = p_ce->ce_dbp_a->put(p_ce->ce_dbp_a, NULL, &key, &data, DB_NODUPDATA);
      if ( (ret == DB_KEYEXIST) || (ret == 0) )
	{
	  /* success */
	  if (ret == 0)
	    {
	      debug_out(dbg_log, "%s: DB_ACCESS: PUT: SUCCESS\n", __FUNCTION__);
	    }
	  else
	    {
	      debug_out(dbg_log, "%s: DB_ACCESS: PUT: DB_KEYEXIST\n", __FUNCTION__);
	    }
	}
      else
	{
	  /* some sort of database error*/
	  debug_out(dbg_log, "%s: DB_ACCESS: PUT: ERROR - failure\n", __FUNCTION__);
	  p_ce->ce_error=1;
	  p_ce->ce_string="Edb";
	  return 1;
	} 
    }
  srand( (unsigned int) r->request_time ^ getpid());
  p_ce->ce_reset_val = (cntval_t) rand();
  p_ce->ce_string = apr_ltoa(r->pool, p_ce->ce_reset_val);
  p_ce->ce_error=0;
  return 1;
 } /* countm_handler_random */

/**
 *@fn countm_handler_updatecounter(struct countent * p_ce, request_rec *r)
 *@brief handle everything else
 *Currently, this function handles everything else, and knows that display_string is next.
 *@return 1 = do not continue handling
 * 0 = continue handling
 * -1 = some sort of error (will not happen for now)
 */
static int
countm_handler_updatecounter(struct countent * p_ce, request_rec *r)
{
  DBT key, data;
  int ret;

  if (p_ce->ce_link_in_db_count == 0 )
    {
      /* link is not in DB_COUNT */
      p_ce->ce_string = "NoLink";
      p_ce->ce_error=0;
      return 1;
    }
  /* attempt to insert host into DB_ACCESS */
  memset(&key,0,sizeof(key));
  key.data = (void *) p_ce->ce_link;
  key.size = strlen(key.data) +1;
  memset(&data,0,sizeof(data));
  data.data = (void *) p_ce->ce_hostname;
  data.size = strlen(p_ce->ce_hostname) + 1;
  debug_out(dbg_log, "%s: DB_ACCESS: PUT: link=%s data=%s\n", __FUNCTION__, (char *)key.data, (char *)data.data);
  ret = p_ce->ce_dbp_a->put(p_ce->ce_dbp_a, NULL, &key, &data, DB_NODUPDATA);
  if ( (ret == DB_KEYEXIST) || (ret == 0) )
    {
      /* success */
      if (ret == 0)
	{
	  debug_out(dbg_log, "%s: DB_ACCESS: PUT: SUCCESS\n", __FUNCTION__);
	}
      else
	{
	  debug_out(dbg_log, "%s: DB_ACCESS: PUT: DB_KEYEXIST\n", __FUNCTION__);
	}
    }
  else
    {
      /* some sort of database error*/
      debug_out(dbg_log, "%s: DB_ACCESS: PUT: ERROR - failure\n", __FUNCTION__);
      p_ce->ce_error=1;
      p_ce->ce_string="Edb";
      return 1;
    } 
  switch(p_ce->ce_count)
    {
    case count_norm_dec:
      if (ret != DB_KEYEXIST) --p_ce->ce_dbc_rec.cnt;
      break;
    case count_inc:
      ++p_ce->ce_dbc_rec.cnt;
      break;
    case count_dec:
      --p_ce->ce_dbc_rec.cnt;
      break;
    case count_notdef:
    case count_norm_inc:
    default:
      if (ret != DB_KEYEXIST) ++p_ce->ce_dbc_rec.cnt;
      break;
    } /* switch */
  p_ce->ce_db_val = p_ce->ce_dbc_rec.cnt;
  /* place new count back into DB_COUNT */
  memset(&key,0,sizeof(key));
  key.data = (void *) p_ce->ce_link;
  key.size = strlen(key.data) +1;
  memset(&data,0,sizeof(data));
  data.data = (void *) &p_ce->ce_dbc_rec;
  data.size = sizeof(p_ce->ce_dbc_rec);
  debug_out(dbg_log, "%s: DB_COUNT: PUT: link=%s: dbc_rec size=%i\n",__FUNCTION__, (char *) key.data, data.size);
  ret = p_ce->ce_dbp_c->put(p_ce->ce_dbp_c, NULL, &key, &data, 0); /* overwrite existing */
  if (ret != 0)
    {
      /* some sort of database error */
      debug_out(dbg_log, "%s: DB_COUNT: PUT: ERROR - failed\n",__FUNCTION__);
      p_ce->ce_error=1;
      p_ce->ce_string="Edb";
      return 1;
    }
  debug_out(dbg_log, "%s: DB_COUNT: PUT: success\n",__FUNCTION__);
  /* return new count */
  debug_out(dbg_log, "%s: DB_COUNT: NEW CNT: %lu\n",__FUNCTION__,p_ce->ce_db_val);
  p_ce->ce_string=apr_ltoa(r->pool, p_ce->ce_db_val);
  p_ce->ce_error=0;
  return 1;
} /* countm_handler_updatecounter */
